#pragma once

#include "B4RDefines.h"
#include <SPI.h>
#include <RH_RF95.h>
#include <RH_ASK.h>

extern void *operator new(size_t size, void *ptr);

namespace B4R {
	typedef void (*SubVoidByteArray)(Array* barray) ;

	//~version: 2.01
	
	//~shortname: ASK
	//~Event: DataAvailable(Data() As Byte)
	//Driver to send and receive adresseable unreliable datagrams via inexpensive ASK (Amplitude Shift Keying) or OOK (On Off Keying) RF transceivers.
	//Example of receiver: http://www.hoperfusa.com/details.jsp?pid=126
	//Example of transmitter: http://www.hoperfusa.com/details.jsp?pid=127
	//Example: <code>
	//Private ask As ASK
	//</code>
	class B4RASK {

		private:      
			uint8_t backend[sizeof(RH_ASK)];
			RH_ASK* ask;
			PollerNode pnode;
			SubVoidByteArray DataAvailableSub;
			Byte buffer[RH_ASK_MAX_PAYLOAD_LEN];
			static void looper(void* b);
		
		public:                  
			//Initializes the module. Returns True if successful.
			//Only one instance of ASK is supported.
			//DataAvailableSub: The sub to call when a datagram is received
			//ThisAddress: Set thisAddres to a number from 0 to 254, only receive data sent to thisAddress. Set ThisAddress to the same number on all units to emulate broadcast(255).
			//Speed: The desired bit rate in bits per second
			//RXpin: The pin that is used to get data from the receiver
			//TXpin: The pin that is used to send data to the transmitter
			//PTTpin: The pin that is connected to the transmitter controller. It will be set HIGH to enable the transmitter (unless pttInverted is true).
			//PTTinverted: True if you desire the pttin to be inverted so that LOW wil enable the transmitter.
			//Example: <code>
			//If ask.Initialize("ask_DataAvailable",45,2000,11,12,10,False) Then
			//  Log("Init OK")
			//Else
			//  Log("Init ERROR")
			//  Return
			//End If
			//</code>
			bool Initialize(SubVoidByteArray DataAvailableSub, byte ThisAddress, int Speed, byte RXpin, byte TXpin, byte PTTpin, bool PTTinverted);

			//Waits until any previous transmit packet is finished being transmitted.
			//Then loads a message into the transmitter and starts the transmitter. 
			//Note that a message length of 0 is NOT permitted. 
			//Data: Array of data to be sent
			//Return true if the message length was valid and it was correctly queued for transmit
			//Example: <code>
			//If ask.Send("Hello") Then Log("Data sent")
			//</code>
			bool Send(ArrayByte* Data);

			//Returns the maximum message length (60).
			//Example: <code>
			//Log(ask.MaxMessageLength)
			//</code>
			UInt getMaxMessageLength();
			
			//Sets the TO address of the next sent message
			//Example: <code>
			//ask.HeaderTo=25
			//</code>
			void setHeaderTo(byte To);

			//Returns the TO address of the last received message
			//Example: <code>
			//Log("HeaderTo: ",ask.HeaderTo)
			//</code>
			byte getHeaderTo();

			//Sets the FROM address of the next sent message
			//Usually it is better to use Initialize2
			//Example: <code>
			//ask.HeaderFrom=63
			//</code>
			void setHeaderFrom(byte From);

			//Returns the From address of the last received message
			//Example: <code>
			//Log("HeaderFrom: ",ask.HeaderFrom)
			//</code>
			byte getHeaderFrom();

			//Sets the ID of the next sent message
			//Example: <code>
			//Dim count as byte
			//If count=255 Then
			//  count=0
			//Else
			//  count=count+1	
			//End If
			//ask.HeaderID=count
			//</code>
			void setHeaderID(byte ID);

			//Returns the ID of the last received message
			//Example: <code>
			//Log("HeaderID: ",ask.HeaderID)
			//</code>
			byte getHeaderID();

			//Sets the FLAGS of the next sent message
			//Set is bitmask of flags to set
			//Clear is bitmask of flags to clear
			//Example: <code>
			//ask.setHeaderFlags(0x80,0)
			//</code>
			void setHeaderFlags(byte Set, byte Clear);

			//Returns the Flags of the last received message
			//Example: <code>
			//Log("HeaderFlags: ",ask.HeaderFlags)
			//</code>
			byte getHeaderFlags();
	};

	
	//~shortname: RF95
	//~Event: DataAvailable(Data() As Byte)
	//Driver to send and receive adresseable unreliable datagrams via a LoRa capable radio transceiver RF95/96/97/98.
	//Example of transceiver: http://www.dragino.com/products/lora/item/126-lora-mini-dev.html
	//Example of transceiver: http://www.dragino.com/products/module/item/102-lora-shield.html
	//Example: <code>
	//Private rf95 As RF95
	//</code>
	
	class B4RRF95 {

		private:      
			uint8_t backend[sizeof(RH_RF95)];
			RH_RF95* rf95;
			PollerNode pnode;
			SubVoidByteArray DataAvailableSub;
			Byte buffer[RH_RF95_MAX_MESSAGE_LEN];
			static void looper(void* b);

		public:                  
			//Initializes the module. Returns True if successful.
			//DataAvailableSub: The sub to call when a datagram is received
			//ThisAddress: Set thisAddres to a number from 0 to 254, only receive data sent to thisAddress. Set ThisAddress to the same number on all units to emulate broadcast(255).
			//Example: <code>
			//If rf95.Initialize("rf95_DataAvailable",45) Then
			//  Log("Init OK")
			//Else
			//  Log("Init ERROR")
			//  Return
			//End If
			//</code>
			bool Initialize(SubVoidByteArray DataAvailableSub, byte ThisAddress);
			
			//Selects one of the 4 predefined modem configurations.
			//Example: <code>
			//rf95.ModemConfig = rf95.Mode_1
			//</code>
			void setModemConfig(RH_RF95::ModemConfigChoice Mode);

			//Sets all the registers required to configure the data modem in the RF95/96/97/98,
			//including the bandwidth, code rate, spreading factor etc. 
			//You can use this to configure the modem with custom configurations if none of the predined configurations in ModemConfigChoice suits you.
			//Reg_1D: Defines bandwidth and coding rate.
			//Reg_1E: Defines spreading factor and CRC
			//Reg_26: Reserved
			//Example: <code>
			//rf95.SetModemRegisters(Bit.Or(rf95.BW_125KHZ,rf95.CODING_RATE_4_8),Bit.Or(rf95.SPREADING_FACTOR_4096CPS,rf95.CRC_PAYLOAD_ON),rf95.REG26_RESERVED)
			//</code>
			void SetModemRegisters(byte Reg_1D, byte Reg_1E, Byte Reg26);
	
			//Asynchronously sends the data.
			//Data: Array of data to be sent
			//Return True if successful.
			//Example: <code>
			//If rf95.Send("Hello") Then Log("Data sent")
			//</code>
			bool Send(ArrayByte* Data);
			
			//Returns the maximum message length (251).
			//Example: <code>
			//Log(rf95.MaxMessageLength)
			//</code>
			UInt getMaxMessageLength();
			
			//Sets the frequency.
			//Example: <code>
			//rf95.Frequency = 868.1
			//</code>
			void setFrequency(float Frequency);
			
			//Sets the transmitter power output level, and configures the transmitter pin.
			//Be a good neighbour and set the lowest power level you need.
			//Some SX1276/77/78/79 and compatible modules (such as RFM95/96/97/98) 
			//use the PA_BOOST transmitter pin for high power output (and optionally the PA_DAC)
			//while some (such as the Modtronix inAir4 and inAir9) 
			//use the RFO transmitter pin for lower power but higher efficiency.
			//You must set the appropriate power level and useRFO argument for your module.
			//Check with your module manufacturer which transmtter pin is used on your module
			//to ensure you are setting useRFO correctly.
			//Failure to do so will result in very low transmitter power output.
			//Caution: legal power limits may apply in certain countries.
			//After init(), the power will be set to 13dBm, with useRFO false (ie PA_BOOST enabled).
			//\param[in] power Transmitter power level in dBm. For RFM95/96/97/98 LORA with useRFO false, 
			//valid values are from +5 to +23.
			//For Modtronix inAir4 and inAir9 with useRFO true (ie RFO pins in use), 
			//valid values are from -1 to 14.
			//\param[in] useRFO If true, enables the use of the RFO transmitter pins instead of
			//the PA_BOOST pin (false). Choose the correct setting for your module.
			//Example: <code>
			//rf95.TxPower(5,False)
			//</code>
			void TxPower(Byte Power, bool useRFO);
			
			//Sets the radio into low-power sleep mode.
			//If successful, the transport will stay in sleep mode until woken by 
			//changing mode to idle, transmit or receive (eg by calling send(), recv(), available() etc)
			//Caution: there is a time penalty as the radio takes a finite time to wake from sleep mode.
			//Example: <code>
			//rf95.Sleep
			//</code>
			bool Sleep();

			//Use the radio's Channel Activity Detect (CAD) function to detect channel activity.
			//Sets the RF95 radio into CAD mode and waits until CAD detection is complete.
			//To be used in a listen-before-talk mechanism (Collision Avoidance) with a reasonable time backoff algorithm.
			//Return true if channel is in use.  
			//Example: <code>
			//If Not(rf95.isChannelActive) Then rf95.Send("Ready")
			//</code>
			bool isChannelActive();

			//Returns the Signal-to-noise ratio (SNR) of the last received message
			//as measured by the receiver in dB.
			//Example: <code>
			//Log("Signal/Noise: ",rf95.LastSNR)
			//</code>
			int getLastSNR();
			
			//Sets the TO address of the next sent message
			//Example: <code>
			//rf95.HeaderTo=25
			//</code>
			void setHeaderTo(byte To);

			//Returns the TO address of the last received message
			//Example: <code>
			//Log("HeaderTo: ",rf95.HeaderTo)
			//</code>
			byte getHeaderTo();

			//Sets the FROM address of the next sent message
			//Usually it is better to use Initialize2
			//Example: <code>
			//rf95.HeaderFrom=63
			//</code>
			void setHeaderFrom(byte From);

			//Returns the From address of the last received message
			//Example: <code>
			//Log("HeaderFrom: ",rf95.HeaderFrom)
			//</code>
			byte getHeaderFrom();

			//Sets the ID of the next sent message
			//Example: <code>
			//Dim count as byte
			//If count=255 Then
			//  count=0
			//Else
			//  count=count+1	
			//End If
			//rf95.HeaderID=count
			//</code>
			void setHeaderID(byte ID);

			//Returns the ID of the last received message
			//Example: <code>
			//Log("HeaderID: ",rf95.HeaderID)
			//</code>
			byte getHeaderID();

			//Sets the FLAGS of the next sent message
			//Set is bitmask of flags to set
			//Clear is bitmask of flags to clear
			//Example: <code>
			//rf95.setHeaderFlags(0x80,0)
			//</code>
			void setHeaderFlags(byte Set, byte Clear);

			//Returns the Flags of the last received message
			//Example: <code>
			//Log("HeaderFlags: ",rf95.HeaderFlags)
			//</code>
			byte getHeaderFlags();
			
			
			//ENUMS
			
			//Default medium range
			//Bandwidth = 125 kHz
			//Code rate = 4/5
			//Spreading factor = 128chips/symbol
			//CRC on.
			static const RH_RF95::ModemConfigChoice Mode_0 = RH_RF95::ModemConfigChoice::Bw125Cr45Sf128;

			//Fast+short range
			//Bandwidth = 500 kHz
			//Code rate = 4/5
			//Spreading factor = 128chips/symbol
			//CRC on.
			static const RH_RF95::ModemConfigChoice Mode_1 = RH_RF95::ModemConfigChoice::Bw500Cr45Sf128;

			//Slow+long range
			//Bandwidth = 31.25 kHz
			//Code rate = 4/8
			//Spreading factor = 512chips/symbol
			//CRC on.
			static const RH_RF95::ModemConfigChoice Mode_2 = RH_RF95::ModemConfigChoice::Bw31_25Cr48Sf512;

			//Slow+long range
			//Bandwidth = 125 kHz
			//Code rate = 4/8
			//Spreading factor = 4096chips/symbol
			//CRC on.
			static const RH_RF95::ModemConfigChoice Mode_3 = RH_RF95::ModemConfigChoice::Bw125Cr48Sf4096;
			
			//REG_1D
			// Bandwidth 7.8 KHz
			static const byte BW_7_8KHZ = 0x00;
			//REG_1D
			// Bandwidth 10.4 KHz
			static const byte BW_10_4KHZ = 0x10;
			//REG_1D
			// Bandwidth 15.6 KHz
			static const byte BW_15_6KHZ = 0x20;
			//REG_1D
			// Bandwidth 20.8 KHz
			static const byte BW_20_8KHZ = 0x30;
			//REG_1D
			// Bandwidth 31,25 KHz
			static const byte BW_31_25KHZ = 0x40;
			//REG_1D
			// Bandwidth 41,7 KHz
			static const byte BW_41_7KHZ = 0x50;
			//REG_1D
			// Bandwidth 62,5 KHz
			static const byte BW_62_5KHZ = 0x60;
			//REG_1D
			// Bandwidth 125 KHz
			static const byte BW_125KHZ = 0x70;
			//REG_1D
			// Bandwidth 250 KHz
			static const byte BW_250KHZ = 0x80;
			//REG_1D
			// Bandwidth 500 KHz
			static const byte BW_500KHZ = 0x90;
			//REG_1D
			// Coding rate 4/5
			static const byte CODING_RATE_4_5 = 0x02;
			//REG_1D
			// Coding rate 4/6
			static const byte CODING_RATE_4_6 = 0x04;
			//REG_1D
			// Coding rate 4/7
			static const byte CODING_RATE_4_7 = 0x06;
			//REG_1D
			// Coding rate 4/8
			static const byte CODING_RATE_4_8 = 0x08;
			
			//REG_1E
			//CRC on
			static const byte CRC_PAYLOAD_ON = 0x04;
			//REG_1E
			//Spreading factor 64 CPS
			static const byte SPREADING_FACTOR_64CPS = 0x60;
			//REG_1E
			//Spreading factor 128 CPS
			static const byte SPREADING_FACTOR_128CPS = 0x70;
			//REG_1E
			//Spreading factor 256 CPS
			static const byte SPREADING_FACTOR_256CPS = 0x80;
			//REG_1E
			//Spreading factor 512 CPS
			static const byte SPREADING_FACTOR_512CPS = 0x90;
			//REG_1E
			//Spreading factor 1024 CPS
			static const byte SPREADING_FACTOR_1024CPS = 0xa0;
			//REG_1E
			//Spreading factor 2048 CPS
			static const byte SPREADING_FACTOR_2048CPS = 0xb0;
			//REG_1E
			//Spreading factor 4096 CPS
			static const byte SPREADING_FACTOR_4096CPS = 0xc0;

			//REG_26
			//Reserved
			static const byte REG26_RESERVED = 0x00;

	};
}
